/*
1  ascending order on sum of the first two elements, use first element as a tie breaker
2 descending order on sum of the first two elements, use first element as a tie breaker
3  ascending order on sum of the first and third elements, use first element as a tie breaker
4  ascending order on maximum of the second and third elements, use first element as a tie breaker
5 descending order on maximum of the second and third elements, use first element as a tie breaker
6  ascending order on minimum of the first and second elements, use third element as a tie breaker
7  ascending order on first element, use second element as a tie breaker
8 descending order on third element, use first element as a tie breaker
9  ascending order on first element, use second element, in descending order, as a tie breaker
10 ascending order on product of the first two elements, use third element as a tie breaker
*/
import java.util.*;
import java.util.function.*;

public class Triples
{
   public static void main(String[] args)
   {
      // Give nice names to some of the lambdas that we need.
      // These types all come from java.util.function
      IntBinaryOperator mul = (x,y) -> x * y;
      BiPredicate<Integer, Integer> lt = (x,y) -> x < y;
      BiPredicate<Integer, Integer> gt = (x,y) -> x > y;

      Triple[] t = {new Triple(3,2,4),new Triple(4,1,9),new Triple(7,6,1),new Triple(7,8,1),
                    new Triple(0,6,4),new Triple(5,9,0),new Triple(4,1,6),new Triple(4,5,5),
                    new Triple(2,8,5),new Triple(5,9,3),new Triple(3,2,9),new Triple(4,2,3)};
      List<Triple> triples = new ArrayList<>(Arrays.asList(t));
      triples.forEach(System.out::print);
      System.out.println("\n");

/*1*/ triples.stream().sorted(new CmpTriples(Triple::e1, Triple::e2, Integer::sum, lt, Triple::e1, lt))
      .forEach(System.out::print);
      System.out.println();

/*2*/ triples.stream().sorted(new CmpTriples(Triple::e1, Triple::e2, Integer::sum, gt, Triple::e1, gt))
      .forEach(System.out::print);
      System.out.println();

/*3*/ triples.stream().sorted(new CmpTriples(Triple::e1, Triple::e3, Integer::sum, lt, Triple::e1, lt))
      .forEach(System.out::print);
      System.out.println();

/*4*/ triples.stream().sorted(new CmpTriples(Triple::e2, Triple::e3, Math::max, lt, Triple::e1, lt))
      .forEach(System.out::print);
      System.out.println();

/*5*/ triples.stream().sorted(new CmpTriples(Triple::e2, Triple::e3, Math::max, gt, Triple::e1, gt))
      .forEach(System.out::print);
      System.out.println();

/*6*/ triples.stream().sorted(new CmpTriples(Triple::e1, Triple::e2, Math::min, lt, Triple::e3, lt))
      .forEach(System.out::print);
      System.out.println();

/*7*/ triples.stream().sorted(new CmpTriples(Triple::e1, Triple::e1, Math::max, lt, Triple::e2, lt))
      .forEach(System.out::print);
      System.out.println();

/*8*/ triples.stream().sorted(new CmpTriples(Triple::e3, Triple::e3, Math::max, gt, Triple::e1, gt))
      .forEach(System.out::print);
      System.out.println();

/*9*/ triples.stream().sorted(new CmpTriples(Triple::e1, Triple::e1, Math::max, lt, Triple::e2, gt))
      .forEach(System.out::print);
      System.out.println();

/*10*/triples.stream().sorted(new CmpTriples(Triple::e1, Triple::e2, mul, lt, Triple::e3, lt))
      .forEach(System.out::print);
      System.out.println();
   }//main()
}//Triples


class CmpTriples implements Comparator<Triple>
{  // These paramaters are all "functional interfaces"
   private ToIntFunction<Triple> s1;
   private ToIntFunction<Triple> s2;
   private IntBinaryOperator reduce;
   private BiPredicate<Integer, Integer> cmp1;
   private ToIntFunction<Triple> s3;
   private BiPredicate<Integer, Integer> cmp2;

   CmpTriples(ToIntFunction<Triple> s1,
              ToIntFunction<Triple> s2,
              IntBinaryOperator reduce,
              BiPredicate<Integer, Integer> cmp1,
              ToIntFunction<Triple> s3,
              BiPredicate<Integer, Integer> cmp2)
   {
      this.s1 = s1;
      this.s2 = s2;
      this.reduce = reduce;
      this.cmp1 = cmp1;
      this.s3 = s3;
      this.cmp2 = cmp2;
   }

   public int compare(Triple t1, Triple t2)
   {  // this is correct only for strict inequalities, < and >
      int n1 = reduce.applyAsInt(s1.applyAsInt(t1), s2.applyAsInt(t1));
      int n2 = reduce.applyAsInt(s1.applyAsInt(t2), s2.applyAsInt(t2));
      int n3 = s3.applyAsInt(t1);
      int n4 = s3.applyAsInt(t2);
      if (cmp1.test(n1, n2)) return -1;
      if (cmp1.test(n2, n1)) return  1;
      if (cmp2.test(n3, n4)) return -1;
      if (cmp2.test(n4, n3)) return  1;
      return 0;
   }
}//CmpTriples


class Triple
{
   private int x;
   private int y;
   private int z;

   public Triple(int x, int y, int z) {
      this.x  = x;
      this.y = y;
      this.z  = z;
   }

   public int e1() { return x; }
   public int e2(){ return y; }
   public int e3() { return z; }

   public String toString() {
      return "(" + x + " " + y + " " + z + ")";
   }
}//Triple
