/*
1  ascending order on sum of the first two elements, use first element as a tie breaker
2 descending order on sum of the first two elements, use first element as a tie breaker
3  ascending order on sum of the first and third elements, use first element as a tie breaker
4  ascending order on maximum of the second and third elements, use first element as a tie breaker
5 descending order on maximum of the second and third elements, use first element as a tie breaker
6  ascending order on minimum of the first and second elements, use third element as a tie breaker
7  ascending order on first element, use second element as a tie breaker
8 descending order on third element, use first element as a tie breaker
9  ascending order on first element, use second element, in descending order, as a tie breaker
10 ascending order on product of the first two elements, use third element as a tie breaker
*/
use std::cmp::Ordering::{Less,Greater,Equal};
use std::cmp::{max,min};

fn main() {
    // selectors
    let e1 = |x: &(i32,i32,i32)| x.0;
    let e2 = |x: &(i32,i32,i32)| x.1;
    let e3 = |x: &(i32,i32,i32)| x.2;
    // reducers
    let add = std::ops::Add::add;
    let mul = std::ops::Mul::mul;
    // comparisons
    let lt  = |x: i32, y: i32| x < y;
    let gt  = |x: i32, y: i32| x > y;

    let mut a = [(3,2,4),(4,1,9),(7,6,1),(7,8,1),(0,6,4),(5,9,0),(4,1,6),(4,5,5),(2,8,5),(5,9,3),(3,2,9),(4,2,3)];
    println!("{:?}\n", a);
  
    // The following line works, but it is NOT a good idea (why?).
    a.sort_by( |x,y| make_cmp(e1, e2, add, lt, e1, lt).compare(x,y) );
    
    // Why doesn't this compile? Why does cmp0.compare need to wrapped in a lambda expression?
    //let cmp0 = make_cmp(e1, e2, add, lt, e1, lt);  a.sort_by( cmp0.compare );  println!("{:?}", a);
    // Answer: Because of compare's signature. It has a first parameter of type &self.

    let cmp0 = make_cmp(e1, e2, add, lt, e1, lt);  a.sort_by(|x,y| cmp0.compare(x,y));  println!("{:?}", a);
    let cmp1 = make_cmp(e1, e2, add, gt, e1, gt);  a.sort_by(|x,y| cmp1.compare(x,y));  println!("{:?}", a);
    let cmp2 = make_cmp(e1, e3, add, lt, e1, lt);  a.sort_by(|x,y| cmp2.compare(x,y));  println!("{:?}", a);
    let cmp3 = make_cmp(e2, e3, max, lt, e1, lt);  a.sort_by(|x,y| cmp3.compare(x,y));  println!("{:?}", a);
    let cmp4 = make_cmp(e2, e3, max, gt, e1, gt);  a.sort_by(|x,y| cmp4.compare(x,y));  println!("{:?}", a);
    let cmp5 = make_cmp(e1, e2, min, lt, e3, lt);  a.sort_by(|x,y| cmp5.compare(x,y));  println!("{:?}", a);
    let cmp6 = make_cmp(e1, e1, max, lt, e2, lt);  a.sort_by(|x,y| cmp6.compare(x,y));  println!("{:?}", a);
    let cmp7 = make_cmp(e3, e3, max, gt, e1, gt);  a.sort_by(|x,y| cmp7.compare(x,y));  println!("{:?}", a);
    let cmp8 = make_cmp(e1, e1, max, lt, e2, gt);  a.sort_by(|x,y| cmp8.compare(x,y));  println!("{:?}", a);
    let cmp9 = make_cmp(e1, e2, mul, lt, e3, lt);  a.sort_by(|x,y| cmp9.compare(x,y));  println!("{:?}", a);
}


struct CmpTriples {
    s1:   fn(&(i32, i32, i32)) -> i32,   // selector 
    s2:   fn(&(i32, i32, i32)) -> i32,   // selector
    r:    fn(i32, i32)         -> i32,   // reducer
    cmp1: fn(i32, i32)         -> bool,  // comparison
    s3:   fn(&(i32, i32, i32)) -> i32,   // selector 
    cmp2: fn(i32, i32)         -> bool   // comparison
}


fn make_cmp(s1:   fn(&(i32, i32, i32)) -> i32,
            s2:   fn(&(i32, i32, i32)) -> i32,
            r:    fn(i32, i32)         -> i32,
            cmp1: fn(i32, i32)         -> bool,
            s3:   fn(&(i32, i32, i32)) -> i32,
            cmp2: fn(i32, i32)         -> bool) -> CmpTriples 
{
    CmpTriples {s1, s2, r, cmp1, s3, cmp2}
}


impl CmpTriples {
    fn compare(&self, x : &(i32,i32,i32), y : &(i32,i32,i32)) -> std::cmp::Ordering 
    {
        let n1 = (self.r)((self.s1)(&x), (self.s2)(&x));
        let n2 = (self.r)((self.s1)(&y), (self.s2)(&y));
        let n3 = (self.s3)(&x);
        let n4 = (self.s3)(&y);
        if      (self.cmp1)(n1, n2) { Less    }
        else if (self.cmp1)(n2, n1) { Greater }
        else if (self.cmp2)(n3, n4) { Less    }
        else if (self.cmp2)(n4, n3) { Greater }
        else                        { Equal   } 
    }
}
