/*
1  ascending order on sum of the first two elements, use first element as a tie breaker
2 descending order on sum of the first two elements, use first element as a tie breaker
3  ascending order on sum of the first and third elements, use first element as a tie breaker
4  ascending order on maximum of the second and third elements, use first element as a tie breaker
5 descending order on maximum of the second and third elements, use first element as a tie breaker
6  ascending order on minimum of the first and second elements, use third element as a tie breaker
7  ascending order on first element, use second element as a tie breaker
8 descending order on third element, use first element as a tie breaker
9  ascending order on first element, use second element, in descending order, as a tie breaker
10 ascending order on product of the first two elements, use third element as a tie breaker
*/
#include <stdio.h>
#include <stdlib.h>

struct triple {
   int x;
   int y;
   int z;
};

// selector functions for struct triple.
int e1(const struct triple *trpl){ return trpl -> x; }
int e2(const struct triple *trpl){ return trpl -> y; }
int e3(const struct triple *trpl){ return trpl -> z; }

void printTriples(const struct triple triples[], int n) {
   for (int i = 0; i < n; i++) {
      printf("(%d %d %d)", triples[i].x, triples[i].y, triples[i].z);
   }
   printf("\n");
}

// Types for making it easier to define struct CmpTriples.
typedef int selector(const struct triple*);
typedef int reducer(int, int);
typedef int comparison(int, int);

struct CmpTriples {
   selector   *s1;
   selector   *s2;
   reducer    *r;
   comparison *cmp1;
   selector   *s3;
   comparison *cmp2;
} cmpTrpl;  // cmpTrpl is a global variable

// This function sets the values in the global structure cmpTrpl.
void make_cmp(selector s1, selector s2, reducer r, comparison cmp1, selector s3, comparison cmp2) {
   cmpTrpl.s1   = s1;
   cmpTrpl.s2   = s2;
   cmpTrpl.r    = r;
   cmpTrpl.cmp1 = cmp1;
   cmpTrpl.s3   = s3;
   cmpTrpl.cmp2 = cmp2;
}

// This is our parameterized comparator function to be used by C's qsort().
//   http://www.cplusplus.com/reference/cstdlib/qsort/
int compar (const void *t1, const void *t2) {
   int n1 = (cmpTrpl.r)((cmpTrpl.s1)(t1), (cmpTrpl.s2)(t1));
   int n2 = (cmpTrpl.r)((cmpTrpl.s1)(t2), (cmpTrpl.s2)(t2));
   int n3 = (cmpTrpl.s3)(t1);
   int n4 = (cmpTrpl.s3)(t2);
   if ((cmpTrpl.cmp1)(n1, n2)) return -1;
   if ((cmpTrpl.cmp1)(n2, n1)) return  1;
   if ((cmpTrpl.cmp2)(n3, n4)) return -1;
   if ((cmpTrpl.cmp2)(n4, n3)) return  1;
   return  0;
}

// reducers
int add(int x, int y){ return x + y; }
int mul(int x, int y){ return x * y; }
int max(int x, int y){ return x >= y ? x : y; }
int min(int x, int y){ return x <= y ? x : y; }
// comparisons
int lt(int x, int y){ return x < y; }
int gt(int x, int y){ return x > y; }

int main()
{
   struct triple triples[] = {{3,2,4},{4,1,9},{7,6,1},{7,8,1},{0,6,4},{5,9,0},
                              {4,1,6},{4,5,5},{2,8,5},{5,9,3},{3,2,9},{4,2,3}};
   printTriples(triples, 12);
   printf("\n");

   int n = 12;
   int size = sizeof(struct triple);

   make_cmp(e1, e2, add, lt, e1, lt);  qsort(triples, n, size, compar);  printTriples(triples, n);
   make_cmp(e1, e2, add, gt, e1, gt);  qsort(triples, n, size, compar);  printTriples(triples, n);
   make_cmp(e1, e3, add, lt, e1, lt);  qsort(triples, n, size, compar);  printTriples(triples, n);
   make_cmp(e2, e3, max, lt, e1, lt);  qsort(triples, n, size, compar);  printTriples(triples, n);
   make_cmp(e2, e3, max, gt, e1, gt);  qsort(triples, n, size, compar);  printTriples(triples, n);
   make_cmp(e1, e2, min, lt, e3, lt);  qsort(triples, n, size, compar);  printTriples(triples, n);
   make_cmp(e1, e1, max, lt, e2, lt);  qsort(triples, n, size, compar);  printTriples(triples, n);
   make_cmp(e3, e3, max, gt, e1, gt);  qsort(triples, n, size, compar);  printTriples(triples, n);
   make_cmp(e1, e1, max, lt, e2, gt);  qsort(triples, n, size, compar);  printTriples(triples, n);
   make_cmp(e1, e2, mul, lt, e3, lt);  qsort(triples, n, size, compar);  printTriples(triples, n);

   return 0;
}
